@@ -6,7 +6,7 @@ use Carp ();
 use IO::Handle ();
 use Exporter 'import';
 
-our $VERSION = '0.003';
+our $VERSION = '0.004';
 
 our @EXPORT = 'cgi';
 
@@ -119,19 +119,26 @@ my %PENDING_CGI;
 
 sub cgi (&) {
   my ($handler) = @_;
-  my $cgi = bless {}, __PACKAGE__;
-  $PENDING_CGI{0+$cgi} = $cgi;
+  my $cgi = bless {pid => $$}, __PACKAGE__;
+  my $cgi_key = 0+$cgi;
+  $PENDING_CGI{$cgi_key} = $cgi; # don't localize, so premature exit can clean up in END
   my ($error, $errored);
   {
     local $@;
-    eval { local $_ = $cgi; $handler->(); die "cgi completed without rendering a response\n" unless $cgi->{headers_rendered}; 1 }
-      or do { $error = $@; $errored = 1 };
+    eval { local $_ = $cgi; $handler->(); 1 } or do { $error = $@; $errored = 1 };
   }
-  _handle_error($cgi, $error) if $errored;
-  delete $PENDING_CGI{0+$cgi};
+  if ($errored) {
+    _handle_error($cgi, $error);
+  } elsif (!$cgi->{headers_rendered}) {
+    _handle_error($cgi, "cgi completed without rendering a response\n");
+  }
+  delete $PENDING_CGI{$cgi_key};
   1;
 }
 
+# cleanup of premature exit, more reliable than potentially doing this in global destruction
+# ModPerl::Registry or CGI::Compile won't run END after each request,
+# but they override exit to throw an exception which we handle already
 END {
   foreach my $key (keys %PENDING_CGI) {
     my $cgi = delete $PENDING_CGI{$key};
@@ -141,6 +148,7 @@ END {
 
 sub _handle_error {
   my ($cgi, $error) = @_;
+  return unless $cgi->{pid} == $$; # in case of fork
   $cgi->{response_status} = "500 $HTTP_STATUS{500}" unless $cgi->{headers_rendered} or defined $cgi->{response_status};
   if (defined(my $handler = $cgi->{on_error})) {
     my ($error_error, $error_errored);
@@ -148,6 +156,7 @@ sub _handle_error {
       local $@;
       eval { $handler->($cgi, $error); 1 } or do { $error_error = $@; $error_errored = 1 };
     }
+    return unless $cgi->{pid} == $$; # in case of fork in error handler
     if ($error_errored) {
       warn "Exception in error handler: $error_error";
       warn "Original error: $error";
@@ -158,7 +167,7 @@ sub _handle_error {
   $cgi->render(text => 'Internal Server Error') unless $cgi->{headers_rendered};
 }
 
-sub set_error_handler { $_[0]{on_error} = $_[1]; $_[0] }
+sub set_error_handler      { $_[0]{on_error} = $_[1]; $_[0] }
 sub set_request_body_limit { $_[0]{request_body_limit} = $_[1]; $_[0] }
 sub set_input_handle       { $_[0]{input_handle} = $_[1]; $_[0] }
 sub set_output_handle      { $_[0]{output_handle} = $_[1]; $_[0] }
@@ -318,32 +327,12 @@ sub body_json {
   return $self->{body_json};
 }
 
-sub set_response_status {
-  my ($self, $status) = @_;
-  if ($self->{headers_rendered}) {
-    Carp::carp "Attempted to set HTTP response status but headers have already been rendered";
-  } else {
-    Carp::croak "Attempted to set unknown HTTP response status $status" unless exists $HTTP_STATUS{$status};
-    $self->{response_status} = "$status $HTTP_STATUS{$status}";
-  }
-  return $self;
-}
-
-sub set_response_content_type {
-  my ($self, $content_type) = @_;
-  if ($self->{headers_rendered}) {
-    Carp::carp "Attempted to set HTTP response content type but headers have already been rendered";
-  } else {
-    $self->{response_content_type} = $content_type;
-  }
-  return $self;
-}
-
 sub add_response_header {
   my ($self, $name, $value) = @_;
   if ($self->{headers_rendered}) {
     Carp::carp "Attempted to add HTTP response header '$name' but headers have already been rendered";
   } else {
+    Carp::croak "Newline characters not allowed in HTTP response header '$name'" if $value =~ tr/\r\n//;
     push @{$self->{response_headers}}, [$name, $value];
   }
   return $self;
@@ -370,13 +359,44 @@ sub add_response_cookie {
     } continue {
       $i += 2;
     }
+    Carp::croak "Newline characters not allowed in HTTP response cookie '$name'" if $cookie_str =~ tr/\r\n//;
     push @{$self->{response_headers}}, ['Set-Cookie', $cookie_str];
   }
   return $self;
 }
 
-sub response_charset     { defined $_[0]{response_charset} ? $_[0]{response_charset} : ($_[0]{response_charset} = 'UTF-8') }
-sub set_response_charset { $_[0]{response_charset} = $_[1]; $_[0] }
+sub set_response_status {
+  my ($self, $status) = @_;
+  if ($self->{headers_rendered}) {
+    Carp::carp "Attempted to set HTTP response status but headers have already been rendered";
+  } else {
+    if ($status =~ m/\A[0-9]+ [^\r\n]*\z/) {
+      $self->{response_status} = $status;
+    } else {
+      Carp::croak "Attempted to set unknown HTTP response status $status" unless exists $HTTP_STATUS{$status};
+      $self->{response_status} = "$status $HTTP_STATUS{$status}";
+    }
+  }
+  return $self;
+}
+
+sub set_response_content_type {
+  my ($self, $content_type) = @_;
+  if ($self->{headers_rendered}) {
+    Carp::carp "Attempted to set HTTP response content type but headers have already been rendered";
+  } else {
+    Carp::croak "Newline characters not allowed in HTTP response content type" if $content_type =~ tr/\r\n//;
+    $self->{response_content_type} = $content_type;
+  }
+  return $self;
+}
+
+sub set_response_charset {
+  my ($self, $charset) = @_;
+  Carp::croak "Space characters not allowed in HTTP response charset" if $charset =~ m/\s/;
+  $self->{response_charset} = $charset;
+  return $self;
+}
 
 sub set_nph {
   my ($self, $value) = @_;
@@ -396,7 +416,8 @@ sub render {
   my ($self, $type, $data) = @_;
   $type = '' unless defined $type;
   Carp::croak "Don't know how to render '$type'" if length $type and !exists $known_types{$type};
-  my $charset = $self->response_charset;
+  my $charset = $self->{response_charset};
+  $charset = 'UTF-8' unless defined $charset;
   my $out_fh = defined $self->{output_handle} ? $self->{output_handle} : *STDOUT;
   if (!$self->{headers_rendered}) {
     my @headers = @{$self->{response_headers} || []};
@@ -408,6 +429,7 @@ sub render {
       $headers_set{lc $name} = 1;
     }
     if (!$headers_set{location} and $type eq 'redirect') {
+      Carp::croak "Newline characters not allowed in HTTP redirect" if $data =~ tr/\r\n//;
       $headers_str = "Location: $data\r\n$headers_str";
     }
     if (!$headers_set{'content-type'} and $type ne 'redirect') {
@@ -495,8 +517,7 @@ CGI::Tiny - Common Gateway Interface, with no frills
     } elsif ($method eq 'POST') {
       $fribble = $cgi->body_param('fribble');
     } else {
-      $cgi->set_response_status(405);
-      $cgi->render;
+      $cgi->set_response_status(405)->render;
       exit;
     }
     die "Invalid fribble parameter" unless length $fribble;
@@ -545,6 +566,13 @@ rendering output.
 
 =back
 
+Most applications are better written in a L<PSGI>-compatible framework (e.g.
+L<Dancer2> or L<Mojolicious>) and deployed in a persistent application server
+so that the application does not have to start up again every time it receives
+a request. CGI::Tiny, and the CGI protocol in general, is only suited for
+restricted deployment environments that can only run CGI scripts, or
+applications that don't need to scale.
+
 See L</"COMPARISON TO CGI.PM">.
 
 This module's interface is currently I<EXPERIMENTAL> and may be changed
@@ -584,6 +612,16 @@ block; use C<exit> to end a CGI script early after rendering a response.
 CGI::Tiny is a minimal interface to the CGI protocol, but can be extended with
 the use of other CPAN modules.
 
+=head2 Fatpacking
+
+L<App::FatPacker> can be used to pack CGI::Tiny, as well as any other pure-perl
+dependencies, into a CGI script so that it can be deployed to other systems
+without having to install the dependencies there. As a bonus, this means the
+script doesn't have to load those modules separately from disk on every
+execution.
+
+  $ fatpack pack script.source.cgi > script.cgi
+
 =head2 JSON
 
 CGI::Tiny has built in support for parsing and rendering JSON content with
@@ -697,8 +735,7 @@ name).
     if (defined(my $match = $routes->match($cgi->path, method => $cgi->method))) {
       $dispatch{$match->name}->($cgi, $match->captures);
     } else {
-      $cgi->set_response_status(404);
-      $cgi->render(text => 'Not Found');
+      $cgi->set_response_status(404)->render(text => 'Not Found');
     }
   };
 
@@ -726,8 +763,12 @@ warned or logged.
 
 Exceptions may occur before or after response headers have been rendered, so
 error handlers should render some response if L</"headers_rendered"> is
-false. If no response has been rendered after the error handler completes, the
-default 500 Internal Server Error response will be rendered.
+false.
+
+If the error handler itself throws an exception, that error and the original
+error will be emitted as a warning. If no response has been rendered after the
+error handler completes or dies, the default 500 Internal Server Error response
+will be rendered.
 
 =head3 set_request_body_limit
 
@@ -929,21 +970,6 @@ L</"set_request_body_limit"> can fit well within the available memory.
 
 =head2 Response
 
-=head3 set_response_status
-
-  $cgi = $cgi->set_response_status(404);
-
-Sets the response HTTP status code. No effect after response headers have been
-rendered. The CGI protocol assumes a status of C<200 OK> if no response status
-is set.
-
-=head3 set_response_content_type
-
-  $cgi = $cgi->set_response_content_type('application/xml');
-
-Sets the response Content-Type header, to override autodetection. No effect
-after response headers have been rendered.
-
 =head3 add_response_header
 
   $cgi = $cgi->add_response_header('Content-Disposition' => 'attachment');
@@ -1031,18 +1057,34 @@ If set to a true value, the cookie will be restricted to HTTPS requests.
 
 =back
 
-=head3 response_charset
+=head3 set_response_status
 
-  my $charset = $cgi->response_charset;
+  $cgi = $cgi->set_response_status(404);
+  $cgi = $cgi->set_response_status('500 Internal Server Error');
 
-Charset to use when rendering C<text>, C<html>, or C<xml> response data,
-defaults to C<UTF-8>.
+Sets the response HTTP status code. No effect after response headers have been
+rendered.
+
+A full status string including a human-readable message will be used as-is. A
+bare status code must be a known
+L<HTTP status code|https://www.iana.org/assignments/http-status-codes/http-status-codes.xhtml>
+and will have the standard human-readable message appended.
+
+The CGI protocol assumes a status of C<200 OK> if no response status is set.
+
+=head3 set_response_content_type
+
+  $cgi = $cgi->set_response_content_type('application/xml');
+
+Sets the response Content-Type header, to override autodetection in
+L</"render">. No effect after response headers have been rendered.
 
 =head3 set_response_charset
 
   $cgi = $cgi->set_response_charset('UTF-8');
 
-Sets L</"response_charset">.
+Set charset to use when rendering C<text>, C<html>, or C<xml> response data,
+defaults to C<UTF-8>.
 
 =head3 set_nph
 
@@ -1084,10 +1126,10 @@ passed in the first call to C<render>, or to C<application/octet-stream> if
 there is no more appropriate value.
 
 C<html>, C<xml>, or C<text> data is expected to be decoded characters, and will
-be encoded according to L</"response_charset">. C<json> data will be encoded to
-UTF-8.
+be encoded according to L</"set_response_charset"> (UTF-8 by default). C<json>
+data will be encoded to UTF-8.
 
-C<redirect> will set a L<Location> header if response headers have not yet been
+C<redirect> will set a C<Location> header if response headers have not yet been
 rendered, and will set a response status of 302 if none has been set by
 L</"set_response_status">. It will not set a C<Content-Type> response header.
 If response headers have already been rendered a warning will be emitted.
@@ -1103,8 +1145,8 @@ The following convenience functions are provided but not exported.
 
   my $date = CGI::Tiny::epoch_to_date $epoch;
 
-Convert a Unix epoch timestamp to a RFC 1123 HTTP date string suitable for use
-in HTTP headers such as C<Date> and C<Expires>.
+Convert a Unix epoch timestamp, such as returned by C<time>, to a RFC 1123 HTTP
+date string suitable for use in HTTP headers such as C<Date> and C<Expires>.
 
 =head2 date_to_epoch
 
@@ -1162,8 +1204,10 @@ L</"set_error_handler">.
 =item *
 
 Request query and body parameter accessors in CGI::Tiny are not context
-sensitive. L</"query_param"> and L</"body_param"> always return a single value,
-and L</"query_param_array"> and L</"body_param_array"> must be used to retrieve
+sensitive, as context sensitivity can lead to surprising behavior and
+L<vulnerabilities|https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-2014-1572>.
+L</"query_param"> and L</"body_param"> always return a single value, and
+L</"query_param_array"> and L</"body_param_array"> must be used to retrieve
 multi-value parameters. CGI::Tiny also does not have a method-sensitive
 C<param> accessor; query or body parameters must be accessed specifically.
 
@@ -1198,7 +1242,7 @@ response before the first call to L</"render">.
 
 =item *
 
-CGI::Tiny does not provide any HTML templating helpers, as this functionality
+CGI::Tiny does not provide any HTML generation helpers, as this functionality
 is much better implemented by other robust implementations on CPAN; see
 L</"Templating">.
 
@@ -1210,6 +1254,39 @@ provided to render appropriate C<Expires> date values.
 
 =back
 
+There are a number of alternatives to CGI.pm but they do not sufficiently
+address the design issues; primarily, none of them gracefully handle
+exceptions or failure to render a response, and several of them have no
+features for rendering responses.
+
+=over
+
+=item *
+
+L<CGI::Simple> shares all of the interface design problems of CGI.pm, though it
+does not reimplement the HTML generation helpers.
+
+=item *
+
+L<CGI::Thin> is ancient and only implements parsing of request query or body
+parameters.
+
+=item *
+
+L<CGI::Minimal> has context-sensitive parameter accessors, and only implements
+parsing of request query/body parameters and uploads.
+
+=item *
+
+L<CGI::Lite> has context-sensitive parameter accessors, and only implements
+parsing of request query/body parameters, uploads, and cookies.
+
+=item *
+
+L<CGI::Easy> has a robust interface, but pre-parses all request information.
+
+=back
+
 =head1 CAVEATS
 
 CGI is an extremely simplistic protocol and relies particularly on the global
@@ -1221,17 +1298,14 @@ CGI::Tiny eschews certain sanity checking for performance reasons. For example,
 Content-Type and other header values set for the response should only contain
 ASCII text with no control characters, but CGI::Tiny does not verify this.
 
-Most applications are better written in a L<PSGI>-compatible framework (e.g.
-L<Dancer2> or L<Mojolicious>) and deployed in a persistent application server
-so that the application does not have to start up again every time it receives
-a request.
-
 =head1 TODO
 
 =over
 
 =item * Uploads/multipart request
 
+=item * Debugging tools
+
 =back
 
 =head1 BUGS
